// -*- C++ -*-
#pragma once

#include "./lib/range.h"
#include "lib/functions.h"
#include "operators.h"
#if __cplusplus >= 202000
#if __has_include(<bit>)
#include <bit>
#endif
#endif


/// @{
/// @addtogroup functions cpu and gpu functions
/// \brief CPU functions
/// In the namespace goopax::extra_operators additional function definitions for CPU data types are provided that otherwise are only provided for gpu data types.

namespace goopax {
namespace impl {

namespace std_poke {
template<typename T, typename enable_if<false, T>::type>
void popcount();
template<typename T, typename enable_if<false, T>::type>
void rotl();
template<typename T, typename enable_if<false, T>::type>
void rotr();
template<typename T, typename enable_if<false, T>::type>
void countl_zero();
}

namespace std_tester {
using namespace std;
using namespace std_poke;
}

template<typename T, typename = void>
struct have_popcount : false_type
{};

template<typename T>
struct have_popcount<T, typename enable_if<sizeof(std_tester::popcount(std::declval<T>())) != 0>::type> : true_type
{};

template<typename T, typename = void>
struct have_rotl : false_type
{};

template<typename T>
struct have_rotl<T, typename enable_if<sizeof(std_tester::rotl(std::declval<T>(), 1)) != 0>::type> : true_type
{};

template<typename T, typename = void>
struct have_rotr : false_type
{};

template<typename T>
struct have_rotr<T, typename enable_if<sizeof(std_tester::rotr(std::declval<T>(), 1)) != 0>::type> : true_type
{};

template<typename T, typename = void>
struct have_countl_zero : false_type
{};

template<typename T>
struct have_countl_zero<T, typename enable_if<sizeof(std_tester::countl_zero(std::declval<T>())) != 0>::type> : true_type
{};


constexpr inline unsigned int popcount32(unsigned int i)
{
    unsigned int ret = 0;
    while (i != 0)
    {
        ret += (i & 1);
        i >>= 1;
    }
    return ret;
}

template<typename T, typename = typename enable_if<is_unsigned<T>::value && (sizeof(T) <= 4)>::type>
constexpr unsigned int popcount_u(T i)
{
    return popcount32(i);
}
constexpr inline unsigned int popcount_u(uint64_t i)
{
    return popcount_u(uint32_t(i)) + popcount_u(uint32_t(i >> 32));
}

template<typename T, typename = typename enable_if<is_integral<T>::value && !have_popcount<unsigned int>::value>::type>
constexpr unsigned int popcount(T i)
{
    /// \ingroup functions
    /// Counts the number of bits in the given value of integral type
    return popcount_u(static_cast<typename make_unsigned<T>::type>(i));
}


template<typename I, typename = typename enable_if<is_integral<I>::value && !have_rotl<unsigned int>::value>::type>
constexpr I rotl(I a, int b)
{
    /// \ingroup functions
    /// rotate left
    typename make_unsigned<I>::type au = a;
    unsigned int                    bu = b;
    bu                                 = bu % (sizeof(typename unrangetype_core<I>::type) * 8);
    return I((au << bu) | (bu == 0 ? 0 : (au >> (sizeof(typename unrangetype_core<I>::type) * 8 - bu))));
}

template<typename I, typename = typename enable_if<is_integral<I>::value && !have_rotr<unsigned int>::value>::type>
constexpr I rotr(I a, int b)
{
    /// \ingroup functions
    /// rotate right
    typename make_unsigned<I>::type au = a;
    unsigned int                    bu = b;
    bu %= (sizeof(typename unrangetype_core<I>::type) * 8);
    return I((au >> bu) | (bu == 0 ? 0 : (au << (sizeof(typename unrangetype_core<I>::type) * 8 - bu))));
}


constexpr inline unsigned int clz32(unsigned int i)
{
    unsigned int ret = 0;
    while ((i & 0x80000000) == 0 && ret < 32)
    {
        ++ret;
        i <<= 1;
    }
    return ret;
}


//@{

template<typename T>
constexpr typename enable_if<(is_integral<T>::value && (sizeof(typename unrangetype_core<T>::type) <= 4) && !have_countl_zero<unsigned int>::value), unsigned int>::type
countl_zero(T i)
{
    /// \ingroup functions
    /// Counts the number of leading zeros in the given value of integral type
    if (i == 0) return 8 * sizeof(typename unrangetype_core<T>::type);
    if (i < 0)
        return clz32((uint32_t)i);
    else
        return clz32((uint32_t)i) - 8 * (4 - sizeof(typename unrangetype_core<T>::type));
}


template<typename T = unsigned int>
constexpr typename enable_if<!have_countl_zero<T>::value, unsigned int>::type
countl_zero(uint64_t i)
{
    /// \ingroup functions
    /// Counts the number of leading zeros in the given value of integral type
    if ((i >> 32) == 0)
        return 32 + clz32(uint32_t(i));
    else
        return clz32(uint32_t(i >> 32));
}

template<typename T>
constexpr typename enable_if<(is_integral<T>::value && !have_countl_zero<unsigned int>::value), unsigned int>::type
countl_one(T i)
{
    /// \ingroup functions
    /// Counts the number of leading ones in the given value of integral type
    return countl_zero(static_cast<T>(~i));
}


template<typename T, typename = typename enable_if<is_integral<T>::value && (sizeof(typename unrangetype_core<T>::type) <= 4)>::type>
[[deprecated("please use countl_zero instead")]] constexpr unsigned int
clz(T i)
{
    /// \ingroup functions
    /// Counts the number of leading zeros in the given value of integral type
    if (i == 0) return 8 * sizeof(typename unrangetype_core<T>::type);
    if (i < 0)
        return clz32((uint32_t)i);
    else
        return clz32((uint32_t)i) - 8 * (4 - sizeof(typename unrangetype_core<T>::type));
}
[[deprecated("please use countl_zero instead")]] constexpr inline unsigned int
clz(uint64_t i)
{
    /// \ingroup functions
    /// Counts the number of leading zeros in the given value of integral type
    if ((i >> 32) == 0)
        return 32 + clz32(uint32_t(i));
    else
        return clz32(uint32_t(i >> 32));
}


//@}

inline double sinpi(double x)
{
    /// \ingroup functions
    /// sin(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::sin(x * PI);
}
inline float sinpi(float x)
{
    /// \ingroup functions
    /// sin(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::sin(x * PI);
}
inline half sinpi(half x)
{
    /// \ingroup functions
    /// sin(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::sin(x * PI);
}
inline double cospi(double x)
{
    /// \ingroup functions
    /// cos(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::cos(x * PI);
}
inline float cospi(float x)
{
    /// \ingroup functions
    /// cos(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::cos(x * PI);
}
inline half cospi(half x)
{
    /// \ingroup functions
    /// cos(x*PI)
    x -= impl::round(x / 2) * 2;
    return impl::cos(x * PI);
}
inline double tanpi(double x)
{
    /// \ingroup functions
    /// tan(x*PI)
    x -= impl::round(x);
    if (x == 0.5)
        return numeric_limits<float>::infinity();
    if (x == -0.5)
        return -numeric_limits<float>::infinity();
    return impl::tan(x * PI);
}
inline float tanpi(float x)
{
    /// \ingroup functions
    /// tan(x*PI)
    x -= impl::round(x);
    if (x == 0.5)
        return numeric_limits<float>::infinity();
    if (x == -0.5)
        return -numeric_limits<float>::infinity();
    return impl::tan(x * PI);
}
inline half tanpi(half x)
{
    /// \ingroup functions
    /// tan(x*PI)
    x -= impl::round(x);
    if (x == 0.5)
        return numeric_limits<float>::infinity();
    if (x == -0.5)
        return -numeric_limits<float>::infinity();
    return impl::tan(x * PI);
}
inline double asinpi(double x)
{
    /// \ingroup functions
    /// asin(x) / PI
    return asin(x) / PI;
}
inline float asinpi(float x)
{
    /// \ingroup functions
    /// asin(x) / PI
    return asin(x) / PI;
}
inline half asinpi(half x)
{
    /// \ingroup functions
    /// asin(x) / PI
    return asin(x) / PI;
}
inline double acospi(double x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return acos(x) / PI;
}
inline float acospi(float x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return acos(x) / PI;
}
inline half acospi(half x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return acos(x) / PI;
}
inline double atanpi(double x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return atan(x) / PI;
}
inline float atanpi(float x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return atan(x) / PI;
}
inline half atanpi(half x)
{
    /// \ingroup functions
    /// acos(x) / PI
    return atan(x) / PI;
}
}

using impl::acos;
using impl::acosh;
using impl::acospi;
using impl::asin;
using impl::asinh;
using impl::asinpi;
using impl::atan;
using impl::atan2;
using impl::atanh;
using impl::atanpi;
using impl::clz;
using impl::cos;
using impl::cosh;
using impl::cospi;
using impl::erf;
using impl::erfc;
using impl::exp;
using impl::exp2;
using impl::log;
using impl::log2;
using impl::sin;
using impl::sinh;
using impl::sinpi;
using impl::tan;
using impl::tanh;
using impl::tanpi;
using impl::tgamma;

}

/// @}
