// -*- C++ -*-

#pragma once

#include "gettype.h"
#include "lib/stdinc.h"
#include "proto.h"
#if __cplusplus >= 202000
#if __has_include(<bit>)
#include <bit>
#endif
#endif

/// \addtogroup reinterpret
/// The reinterpret functions provide functionalities for bit-wise conversion between data types without changing the underlying representation.
/// Various types can be converted into each other: intrinsic types to goopax structs, goopax structs to std::vectors of other goopax structs, vectors to intrinsic types, etc.
/// The sizes of the source value and the destination value must match, otherwise the compiler will report an error. If the source type is an std::vector, the size is checked at runtime.
/// If the destination type is an std::vector, its size is chosen so that it matches the source type.
/// If the source type is an CPU type, the destination type must also be a CPU type. If the source type is a GPU type, the destination type must also be a GPU type.
/// \{

namespace goopax {
namespace impl {


namespace std_poke {
template<typename T, typename enable_if<false, T>::type>
void bit_cast();
}


namespace std_tester {
using namespace std;
using namespace std_poke;
}


template<typename TO, typename FROM, typename = void>
struct have_bit_cast : false_type
{};

template<typename TO, typename FROM>
struct have_bit_cast<TO, FROM, typename enable_if<sizeof(std_tester::bit_cast<TO>(std::declval<FROM>())) == sizeof(TO)>::type> : true_type
{};


/**
       \brief bit-wise conversion
       \ingroup reinterpret
	Reinterpret bits into another type.
	\param TO destination type
	\param from source
	This function reinterprets the input parameter into another type.
	FROM and TO must have the same size. They may be CPU type, GPU type, std::vector, or goopax struct type.
	
	Example: 
	gpu_double a = 17;
	std::array<gpu_int, 2> b = reinterpret<std::array<gpu_int, 2> >(a);
	gpu_uint64 c = reinterpret<gpu_uint64> c;
       */


// this is constexpr in C++20, if std::bit_cast available
template<class TO, class FROM>
constexpr inline typename enable_if<(std::is_trivially_copyable<TO>::value && std::is_trivially_copyable<FROM>::value && sizeof(TO) == sizeof(FROM) && have_bit_cast<TO, FROM>::value),
                                    TO>::type
reinterpret(const FROM& from);

template<class TO, class FROM>
inline typename enable_if<(std::is_trivially_copyable<TO>::value && std::is_trivially_copyable<FROM>::value && sizeof(TO) == sizeof(FROM) && !have_bit_cast<TO, FROM>::value),
                          TO>::type
reinterpret(const FROM& from);


/**
       \brief Reinterpreting buffer value type
       \ingroup reinterpret
       Changes the buffer value type without changing the data content. The returned buffer points to the same memory buffer as the original buffer.
       \param TO destination type
       \param from source buffer
    */
template<typename TO, typename TFROM, typename SIZE_FROM>
inline TO reinterpret(const buffer<TFROM, SIZE_FROM>& from);


template<class TO, class FROM>
inline typename enable_if<(!std::is_trivially_copyable<TO>::value || !std::is_trivially_copyable<FROM>::value),
                          TO>::type
reinterpret(const FROM& from);

namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_reinterpret_gpuvec(gpu_writer_base_impl** to, gpu_writer_base_impl** from, size_t Nfrom) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) void goopax_4_reinterpret_struct(const gpu_struct::mapframe_base&, const gpu_struct::mapframe_base&, const gpu_struct::mapframe_base&, const gpu_struct::mapframe_base&, const void*, void*, bool is_gpu) noexcept;
}
}


template<class TO, class FROM>
struct debug_ptr_check
{};

template<class TO, class FROM, typename TA>
struct debug_ptr_check<checknan<TO, TA>*, checknan<FROM, TA>*>
{
    static_assert(sizeof(TO) == sizeof(FROM), "Casting between debug pointers of different base type size is not allowed.");
};

template<class TO, class FROM, size_t NTO, size_t NFROM, memory::scope memscope>
void reinterpret_switch3(array<gpu_type<TO, memscope>, NTO>& to, const array<gpu_type<FROM, memscope>, NFROM>& from)
{
    debug_ptr_check<TO, FROM> ptrcheck;
    (void)ptrcheck;
    static_assert(gettype<TO>::size * NTO == gettype<FROM>::size * NFROM, "Reinterpret: Objects have different size.");
    vector<gpu_writer_base_impl*> toi(NTO);
    for (size_t k = 0; k < NTO; ++k)
    {
        toi[k] = to[k].impl.get();
    }
    vector<gpu_writer_base_impl*> fromi(NFROM);
    for (size_t k = 0; k < NFROM; ++k)
    {
        fromi[k] = from[k].impl.get();
    }
    check_errno(interface::goopax_4E_reinterpret_gpuvec(&toi[0], &fromi[0], NFROM));
}


template<class F, class FA, size_t NF, class T, class TA, size_t NT>
void reinterpret_switch3(array<checknan<T, TA>, NT>& to, const array<checknan<F, FA>, NF>& from, typename std::enable_if<NF == 1 || NT == 1, int>::type = 0)
{
    size_t num_uninit = 0;
    for (size_t k = 0; k < NF; ++k)
    {
        if (from[k].is_uninitialized())
            ++num_uninit;
    }
    if (num_uninit == 0)
    {
        array<F, NF> fromu;
        for (size_t k = 0; k < NF; ++k)
        {
            fromu[k] = from[k];
        }
        const auto tou = reinterpret<array<T, NT>>(fromu);
        for (size_t k = 0; k < NT; ++k)
        {
            to[k] = tou[k];
        }
    }
    else if (num_uninit == NF)
        to = array<checknan<T, TA>, NT>();
    else
    {
        warning() << "GOOPAX: reinterpret: " << num_uninit << " of " << NF << " elements are uninitialized!" << endl;
        abort();
    }
}

template<class F, size_t NF, class T, class TA, size_t NT>
void reinterpret_switch3(array<checknan<T, TA>, NT>& to, const array<F, NF>& from, typename enable_if<(NF == 1 || NT == 1) && (is_fundamental<F>::value || is_pointer<F>::value), int>::type = 0)
{
    array<T, NT> tmp = reinterpret<array<T, NT>>(from);
    std::copy(tmp.begin(), tmp.end(), to.begin());
}

template<class F, class FA, size_t NF, class T, size_t NT>
void reinterpret_switch3(array<T, NT>& to, const array<checknan<F, FA>, NF>& from, typename enable_if<(NF == 1 || NT == 1) && (is_fundamental<T>::value || is_pointer<T>::value), int>::type = 0)
{
    array<F, NF> tmp;
    std::copy(from.begin(), from.end(), tmp.begin());
    to = reinterpret<array<T, NT>>(tmp);
}


template<class TO, class FROM>
inline void reinterpret_switch3(TO& to, const FROM& from)
{
    static_assert(get_size<TO>::value == get_size<FROM>::value, "Reinterpret: Objects have different size.");
    static_assert(is_same<typename unrangetype<typename get_X_type<FROM>::type>::type, typename unrangetype<typename get_X_type<TO>::type>::type>::value, "");
    const auto& mf_from  = gpu_struct::mapframe<FROM>::get_static_mf();
    const auto& mf_fromu = gpu_struct::mapframe<typename goopax_struct_changetype<FROM, unsigned int>::type>::get_static_mf();
    const auto& mf_to    = gpu_struct::mapframe<TO>::get_static_mf();
    const auto& mf_tou   = gpu_struct::mapframe<typename goopax_struct_changetype<TO, unsigned int>::type>::get_static_mf();

    interface::goopax_4_reinterpret_struct(mf_from, mf_fromu, mf_to, mf_tou, &from, &to, is_same<TO, typename gettype<TO>::gpu>::value);
}


template<class TO, class FROM>
inline void reinterpret_switch3(std::vector<TO>& to, const std::vector<FROM>& from)
{
    static_assert((gettype<TO>::size % gettype<FROM>::size == 0) || (gettype<FROM>::size % gettype<TO>::size == 0), "Reinterpret: Object sizes must be divisible.");

    to.resize(from.size() * gettype<FROM>::size / gettype<TO>::size);
    assert(from.size() * gettype<FROM>::size == to.size() * gettype<TO>::size); 

    constexpr size_t bsto   = gettype<FROM>::size / gettype<TO>::size == 0 ? 1 : gettype<FROM>::size / gettype<TO>::size;
    constexpr size_t bsfrom = gettype<TO>::size / gettype<FROM>::size == 0 ? 1 : gettype<TO>::size / gettype<FROM>::size;

    auto p_to   = to.begin();
    auto p_from = from.begin();
    while (p_to != to.end())
    {
        array<FROM, bsfrom> tmp;
        std::copy(p_from, p_from + bsfrom, tmp.begin());
        array<TO, bsto> bto = reinterpret<array<TO, bsto>>(tmp);
        for (size_t k = 0; k < bsto; ++k)
        {
            p_to[k] = bto[k];
        }
        p_to += bsto;
        p_from += bsfrom;
    }
#if GOOPAX_DEBUGHARD >= 2
    assert(p_to == to.end());
    assert(p_from == from.end());
#endif
}

template<typename TO, typename FROM>
inline void reinterpret_switch3(TO& to, const std::vector<FROM>& from)
{
    static_assert(gettype<TO>::size % gettype<FROM>::size == 0, "Reinterpret: Object sizes must be divisible.");
    array<FROM, (gettype<TO>::size / gettype<FROM>::size)> tmp;
    std::copy(from.begin(), from.end(), tmp.begin());
    to = reinterpret<TO>(tmp);
}

template<typename TO, typename FROM>
inline void reinterpret_switch3(std::vector<TO>& to, const FROM& from)
{
    static_assert(gettype<FROM>::size % gettype<TO>::size == 0, "Reinterpret: Object sizes must be divisible.");
    auto tmp = reinterpret<array<TO, (gettype<FROM>::size / gettype<TO>::size)>>(from);
    to.assign(tmp.begin(), tmp.end());
}


template<typename TO, typename FROM>
inline void reinterpret_switch2(TO& to, const FROM& from)
{
    reinterpret_switch3(to, array<FROM, 1>({ { from } }));
}

template<typename TO, typename FROM>
inline void reinterpret_switch2(TO& to, const std::vector<FROM>& from)
{
    reinterpret_switch3(to, from);
}

template<typename TO, typename FROM, size_t NFROM>
inline void reinterpret_switch2(TO& to, const goopax::impl::array<FROM, NFROM>& from)
{
    reinterpret_switch3(to, from);
}


template<typename TO, typename FROM>
inline void reinterpret_switch(TO& to, const FROM& from)
{
    array<TO, 1> tmp;
    reinterpret_switch2(tmp, from);
    to = tmp[0];
}

template<typename TO, typename FROM>
inline void reinterpret_switch(std::vector<TO>& to, const FROM& from)
{
    reinterpret_switch2(to, from);
}


template<typename TO, typename FROM, size_t NTO>
inline void reinterpret_switch(goopax::impl::array<TO, NTO>& to, const FROM& from)
{
    reinterpret_switch2(to, from);
}


// std::bit_cast from C++-20 will be used if available.
template<class TO, class FROM>
constexpr inline typename enable_if<(std::is_trivially_copyable<TO>::value && std::is_trivially_copyable<FROM>::value && sizeof(TO) == sizeof(FROM) && have_bit_cast<TO, FROM>::value),
                                    TO>::type
reinterpret(const FROM& from)
{
    static_assert(is_same<TO, typename unrangetype<TO>::type>::value, "");
    static_assert(is_same<FROM, typename unrangetype<FROM>::type>::value, "");

    return std_tester::bit_cast<TO>(from);
}

template<class TO, class FROM>
inline typename enable_if<(std::is_trivially_copyable<TO>::value && std::is_trivially_copyable<FROM>::value && sizeof(TO) == sizeof(FROM) && !have_bit_cast<TO, FROM>::value),
                          TO>::type
reinterpret(const FROM& from)
{
    static_assert(is_same<TO, typename unrangetype<TO>::type>::value, "");
    static_assert(is_same<FROM, typename unrangetype<FROM>::type>::value, "");

    TO to;
    std::memcpy(&to, &from, sizeof(FROM));
    return to;
}


template<class TO, class FROM>
inline typename enable_if<(!std::is_trivially_copyable<TO>::value || !std::is_trivially_copyable<FROM>::value),
                          TO>::type
reinterpret(const FROM& from)
{
    TO to;
    reinterpret_switch(to, from);
    return to;
}


template<typename TO, typename TFROM, typename SIZE_FROM>
inline TO reinterpret(const buffer<TFROM, SIZE_FROM>& from)
{
    TO to(*from.get_buf(), INITTEX_REINTERPRET());
    return to;
}
}
using impl::reinterpret;
}

/// \}
