// -*- C++ -*-
// @@@ CONVERT_TYPES_IGNORE @@@
#pragma once
#include "../goopax_bits/lib/range.h"
#include "../goopax_bits/lib/stdinc.h"
#include <map>
#include <regex>


namespace goopax {

namespace impl {

using std::map;

struct parambase;

template<typename = void>
map<string, parambase*>& parammap()
{
    static map<string, parambase*> parammap;
    return parammap;
}


struct parambase
{
    const string name;
    const string description;
    bool         seen                  = false;
    virtual void read(const string& s) = 0;

    virtual void out(ostream& s) const = 0;
    virtual void missing() const
    {
        throw std::runtime_error("missing parameter '" + name + "'");
    }

    parambase(const string& name0, const string& description0)
        : name(name0)
        , description(description0)
    {
        if (parammap().find(name) != parammap().end())
        {
            stringstream ss;
            ss << "Parameter '" << name0 << "' defined twice!" << endl;
            string s = ss.str();
            printf("%s", s.c_str());
            throw std::runtime_error(s);
        }
        parammap().insert({ name, this });
    }

    parambase(const parambase&) = delete;
    parambase& operator=(const parambase&) = delete;

    virtual ~parambase()
    {
        parammap().erase(name);
    }
};


inline void param_readhelp(string& t, const string& s)
{
    t = s;
}

template<typename T>
void param_readhelp(T& t, const string& s)
{
    istringstream ss(s);
    const auto    oldex = ss.exceptions();
    ss.exceptions(ios::badbit | ios::failbit);
    bool ok = (bool)(ss >> t);
    if (!ok)
    {
        throw std::runtime_error("Error converting string '" + s + "' to " + pretty_typename(typeid(T)));
    }
    ss.exceptions(oldex);
    {
        int c = ss.get();
        (void)c;

        while (!ss.eof())
        {
            if (c == ' ')
                cerr << "param: Got trailing space." << endl;
            else if (c == '\n')
                cerr << "param: Got trailing newline." << endl;
            else
            {
                stringstream ss;
                if (isprint(c))
                    ss << " [" << (char)c << "]";
                cerr << "Got trailing character. c=" << c << ss.str() << " while converting string '" << s << " to type " << pretty_typename(typeid(T));
                throw std::runtime_error("Got trailing character");
            }
            c = ss.get();
        }
    }
}

template<class T>
struct refbase
{
    private:
    T data;

    public:
    operator const T&() const
    {
        return data;
    }
    operator T&()
    {
        return data;
    }
};


template<class T>
struct PARAM : public parambase
    , public conditional<is_fundamental<T>::value, refbase<T>, T>::type
{
    virtual void read(const string& s) final override
    {
        param_readhelp(static_cast<T&>(*this), s);
    }

    inline const T& operator()() const
    {
        return *this;
    }
    PARAM(const string& name, const string& description = "")
        : parambase(name, description)
    {
    }


    PARAM& operator=(const T& t)
    {
        static_cast<T&>(*this) = t;
        return *this;
    }

    virtual ~PARAM()
    {}

    virtual void out(ostream& s) const override
    {
        s << pretty_typename(typeid(T));
    }
};


template<class T>
struct PARAMOPT : public PARAM<T>
{
    virtual void out(ostream& s) const final override
    {
        PARAM<T>::out(s);
        s << " [" << static_cast<const T&>(*this) << "]";
    }

    virtual void missing() const final override
    {}

    PARAMOPT& operator=(const T& t)
    {
        PARAM<T>::operator=(t);
        return *this;
    }

    PARAMOPT(const string& name, const T& def, const string& description = "")
        : PARAM<T>(name, description)
    {
        static_cast<T&>(*this) = def;
    }
};


template<typename = void>
void add_param(const std::string& name, const std::string& value)
{
    auto p = parammap().find(name);
    if (p != parammap().end())
    {
        p->second->read(value);
        if (p->second->seen != false)
        {
            cerr << "Parameter named '" << name << "' used more than once!" << endl;
        }
        p->second->seen = true;
    }
    else
    {
        throw std::runtime_error("Unknown parameter '" + name + "'");
    }
}

template<typename = void>
inline void init_params(int& argc, char**& argv)
{
    /// \ingroup device
    /// May be called to allow passing command line arguments to GOOPAX or to PARAM or PARAMOPT user values.
    /// \param argc argc from the main() function
    /// \param argv argv from the main() function
    /// \param paramfile optional filename of configuration file


    char** src  = argv + 1;
    char** dest = argv + 1;
    char** end  = argv + argc;

    while (src != end)
    {
        std::smatch what;

        std::string a = *src;
        if (std::regex_match(a, what, std::regex("--([^=]*)=(.*)")))
        {
            add_param(what[1], what[2]);

            ++src;
            --argc;
        }
        else if (a == "--help")
        {
            cout << "Usage: " << argv[0] << " [--help] [--<OPTION>=<VALUE> ...]" << endl
                 << endl
                 << "options:" << endl;

            for (auto& p : parammap())
            {
                cout << p.first << " : ";
                p.second->out(cout);
                cout << endl;
            }
            exit(0);
        }
        else
        {
            *dest++ = *src++;
        }
    }

    for (auto& p : parammap())
    {
        if (!p.second->seen)
        {
            p.second->missing();
        }
    }
}
}
using impl::init_params;
using impl::PARAM;
using impl::PARAMOPT;
}
